import os
import time
import argparse
from datetime import datetime
from functools import partial
from concurrent.futures import ProcessPoolExecutor
from multiprocessing import freeze_support
import utils

logger = utils.logger


def download_videos(index_post, skip_existing, notify, download_photos, zip_photos, description, add_date, cookie, add_username, convert_video, delete_original, embed_metadata, save_metadata, keep_corrupt, total_videos, set_cookie=None):
    index, post = index_post
    parsed_post = utils.parse_json(post)
    if not parsed_post:
        return

    filename = utils.generate_filename(parsed_post["video_id"], parsed_post["username"], parsed_post["timestamp"], parsed_post["title"], description, add_date, add_username)

    if parsed_post["is_image_post"]:
        if not os.path.exists(f'{filename}.zip' if zip_photos else filename):
            if download_photos:
                if utils.download_photos(filename, parsed_post["video_id"], parsed_post["image_urls"], parsed_post["music_title"], parsed_post["music_url"], parsed_post["is_aweme"], parsed_post["is_story"], parsed_post["is_story_note"], parsed_post["share_url"], post, cookie, save_metadata, zip_photos):
                    print(f"[{index}/{total_videos}]")
            else:
                logger.info("SKIPPING PHOTO POST: %s", parsed_post["video_id"])
        return

    video, sd_video = f"{filename}.mp4", f"{filename}~sd.mp4"

    if os.path.isfile(video):
        return

    if os.path.isfile(sd_video) and skip_existing:
        logger.info("SD version of %s already exists. Skipping.", filename)
        return
        
    source_video = utils.get_source_encode(parsed_post["video_id"])

    if source_video:
        parsed_post["is_sd"] = source_video["size"] == parsed_post["sd_size"]
        if not parsed_post["is_sd"]:
            parsed_post["video_urls"] = [source_video["play_url"]]
            parsed_post["file_hash"] = None
    
    if os.path.isfile(sd_video) and parsed_post["is_sd"]:
        logger.info("Source quality video is not available and SD version of %s already exists. Skipping.", filename)
        return

    if not parsed_post["is_aweme"] and not set_cookie and parsed_post["is_sd"]:
        post = utils.get_post_info(parsed_post["video_id"], request_type="web", cookie=cookie)      
        if not post or not post.get("post_info"):
            return
            
        post, set_cookie = post.get('post_info'), post.get('set_cookie')
        parsed_post = utils.parse_json(post)
        if not parsed_post:
            return
            
    if notify:
            show_notification(parsed_post["username"], parsed_post["video_id"])

    if save_metadata:
            utils.dump_metadata(filename, post)

    video = sd_video if parsed_post["is_sd"] else video
    if utils.download_video(video, parsed_post["video_id"], parsed_post["video_urls"], parsed_post["file_hash"], cookie, set_cookie, keep_corrupt) != 1:
        if not parsed_post["is_sd"]:
            if not utils.check_source_encode(video, sd_video):
                video=sd_video
        if embed_metadata:
            utils.add_metadata(video, parsed_post)
        if convert_video:
            utils.convert_to_h264(video, delete_original)
        print(f"[{index}/{total_videos}]")

def show_notification(username, video_id):
    message = f"{username} uploaded a new video: {video_id}"
    if os.name == 'nt':
        os.system(f'msg.exe * /TIME:1 {message}')
    else:
        os.system(f'notify-send "{message}"')

def validate_date(date_string, date_format='%Y-%m-%d'):
    try:
        datetime.strptime(date_string, date_format)
        return True
    except ValueError:
        return False

def scrape(user, download_dir, amount_time, description, info, add_date, skip_existing, notify, 
    max_downloads, amount_videos, download_photos, zip_photos, stories, cookie, add_username, as_user, 
    convert_video, delete_original, embed_metadata, save_metadata, keep_corrupt, likes, reposts, request_base):

    base_directory = os.getcwd()
    
    user_info = utils.get_user_info(user, cookie)        
    if not user_info:
        print()
        return
   
    username = user_info.get('uniqueId')
    user_id = user_info.get('id')
    sec_uid = user_info.get('secUid')
    avatar_url = user_info.get('avatarLarger')
    private_account = user_info.get('privateAccount')
   
    if info:
        logger.info('USER ID: %s', user_id)
        logger.info('SEC UID: %s', sec_uid)
        logger.info('CURRENT USERNAME: %s', username)
        return

    logger.info('DOWNLOADING: %s (uid:%s)', username, user_id)

    if as_user:
        download_dir = os.path.join(os.getcwd(), username)

    download_dir = download_dir or os.path.join(os.getcwd(), user)
    os.makedirs(download_dir, exist_ok=True)
    os.chdir(download_dir)

    if not os.path.exists('user_info.json'):
        utils.save_user_info(user_info)

    if avatar_url:
        utils.get_avatar(avatar_url)

    if private_account:
        logger.info("Profile is private")
    
    if stories:
        try:
            story_ids = utils.get_video_ids(sec_uid, user_id, amount_time, amount_videos=None, request_type='aweme_stories', cookie=cookie)
            total_stories = len(story_ids)

            if total_stories < 1:
                logger.info('NO STORIES FOUND')
            else:
                logger.info('FOUND %s STORIES', total_stories)
                os.makedirs('Stories', exist_ok=True)
                os.chdir('Stories')

                partial_download = partial(
                    download_videos,
                    skip_existing=skip_existing,
                    notify=notify,
                    download_photos=download_photos,
                    zip_photos=zip_photos,
                    description=description,
                    add_date=add_date,
                    cookie=cookie,
                    add_username=add_username,
                    convert_video=convert_video,
                    delete_original=delete_original,
                    embed_metadata=embed_metadata,
                    save_metadata=save_metadata,
                    keep_corrupt=keep_corrupt,
                    total_videos=total_stories)

                with ProcessPoolExecutor(max_workers=max_downloads) as executor:
                    executor.map(partial_download, enumerate(story_ids, 1))
        except Exception:
            logger.exception('ERROR DOWNLOADING STORIES: %s.', username)
        finally:
            os.chdir(download_dir)
    
    request_type = f"{request_base}_posts"
    
    if likes:
        os.makedirs('Likes', exist_ok=True)
        os.chdir('Likes')
        request_type = f"{request_base}_likes"
    
    if reposts:
        os.makedirs('Reposts', exist_ok=True)
        os.chdir('Reposts')
        request_type = 'web_reposts'

    try:
        video_ids = utils.get_video_ids(sec_uid, user_id, amount_time, amount_videos, request_type, cookie)
        total_videos = len(video_ids)

        if total_videos < 1:
            logger.info('NO VIDEO IDS FOUND FOR: %s\n', username)
            return

        logger.info('DOWNLOADING %s VIDEOS', total_videos)        

        partial_download = partial(
            download_videos,
            skip_existing=skip_existing,
            notify=notify,
            download_photos=download_photos,
            zip_photos=zip_photos,
            description=description,
            add_date=add_date,
            cookie=cookie,
            add_username=add_username,
            convert_video=convert_video,
            delete_original=delete_original,
            embed_metadata=embed_metadata,
            save_metadata=save_metadata,
            keep_corrupt=keep_corrupt,
            total_videos=total_videos)

        with ProcessPoolExecutor(max_workers=max_downloads) as executor:
            executor.map(partial_download, enumerate(video_ids, 1))

        logger.info('FINISHED DOWNLOADING: %s\n', username)
    except Exception:
        logger.exception('ERROR DOWNLOADING: %s.', username)
    finally:
        os.chdir(base_directory)
    return


def main():
    parser = argparse.ArgumentParser(
        prog='Awemer', description='Scrape TikTok profiles.')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-u', '--user', nargs='+',
                        help='Username, user id, secID, or text file with usernames/userIDs one per line.')
    group.add_argument('-v', '--video', default=None, 
                        help='Video ID or TikTok link to the post to download.')
    parser.add_argument('-d', '--download_dir', default=None,
                        help='Directory to download files to.')
    parser.add_argument('-t', '--date', default=None,
                        help='Date until which to download videos in YYYY-MM-DD format. If none specified it will download all videos.')
    parser.add_argument('-s', '--description', action='store_true',
                        default=False, help='Add video description to the filename.')
    parser.add_argument('-i', '--info', action='store_true',
                        default=False, help='Print userID and secID.')
    parser.add_argument('-ad', '--add_date', action='store_true', default=False,
                        help='Add video date at the beginning of the filename.')
    parser.add_argument('-se', '--skip_existing', action='store_true',
                        default=False, help='Skip existing sd videos.')
    parser.add_argument('-l', '--loop', type=int, default=None,
                        help='Run scraper in loop, wait X seconds between iterations')
    parser.add_argument('-n', '--notify', action='store_true',
                        default=False, help='Notify when user has upload a new video.')
    parser.add_argument('-md', '--max_downloads', type=int, default=1,
                        help='Maximum amount of simultaneous downloads. Default=1.')
    parser.add_argument('-mv', '--max_videos', type=int, default=None,
                        help='Maximum amount of videos to download.')
    parser.add_argument('-dp', '--download_photos', action='store_true',
                        default=False, help='Download photo posts.')
    parser.add_argument('-zp', '--zip_photos', action='store_true',
                        default=False, help='Download photo posts to a zip file.')
    parser.add_argument('-st', '--stories', action='store_true',
                        default=False, help='Download user stories.')
    parser.add_argument('-c', '--cookie', default=None,
                        help='sid_tt/sessionid_ss value.')
    parser.add_argument('-au', '--add_username', action='store_true',
                        default=False, help='Add username at the beginning of the filename.')
    parser.add_argument('--as_user', action='store_true', default=False,
                        help='Download videos to a folder named as the current username.')
    parser.add_argument('--convert', action='store_true', default=False,
                        help='Convert downloaded HEVC video to h264.')
    parser.add_argument('--delete_orig', action='store_true', default=False,
                        help='Delete original HEVC video when converted to h264.')
    parser.add_argument('-em', '--embed_metadata', action='store_true', default=False,
                        help='Embed metadata to the video.')
    parser.add_argument('-sm', '--save_metadata', action='store_true', default=False,
                        help='Save metadata to a JSON file.')
    parser.add_argument('-kc', '--keep_corrupt', action='store_true', default=False,
                        help='Keep videos even if they fail integrity check.')
    parser.add_argument('--likes', action='store_true', default=False,
                        help="Download user's liked videos.")
    parser.add_argument('--reposts', action='store_true', default=False,
                        help="Download user's reposted videos.")
    parser.add_argument(
    '--api', choices=['1', '2'], default='1',
                        help='Api used to fetch video IDs and details: 1 for "TikTok Web", 2 for "TikTok Mobile". Default=1.'
)

    args = parser.parse_args()
    
    if args.stories and not args.cookie:
        parser.error('--stories only works if --cookie (-c) is supplied.')
    if args.api == '2' and args.user and not args.cookie:
        parser.error('--api 2 (TikTok Mobile) only works if --cookie (-c) is supplied.')
    if args.date and not validate_date(args.date):
        parser.error('Invalid date format. Use YYYY-MM-DD.')
    if args.delete_orig and not args.convert:
        parser.error('--delete_orig only works if --convert is supplied.')
    if args.as_user and args.download_dir:
        parser.error('argument --download_dir (-d) not allowed with argument --as_user .')
    if args.likes and args.reposts:
        parser.error('argument --likes not allowed with argument --reposts .')
    if args.zip_photos:
        args.download_photos = True
        
    base_mapping = {
        '1': 'web',
        '2': 'aweme'
    }

    request_base = base_mapping[args.api]
    
    cookie = f'sid_tt={args.cookie}; sessionid={args.cookie}; sessionid_ss={args.cookie}' if args.cookie else None

    def start_scrape():
        if args.user:
            for user in args.user:
                if os.path.isfile(user):
                    args.user.remove(user)
                    with open(user, 'r', encoding="utf-8") as profiles:
                        for line in profiles:
                            if line.strip():
                                args.user.append(line.strip())

            for user in args.user:
                scrape(user, args.download_dir, args.date, args.description, args.info, args.add_date, args.skip_existing,
                       args.notify, args.max_downloads, args.max_videos, args.download_photos, args.zip_photos,
                       args.stories, cookie, args.add_username, args.as_user, args.convert, args.delete_orig,
                       args.embed_metadata, args.save_metadata, args.keep_corrupt, args.likes, args.reposts, request_base)

        elif args.video:
            post = utils.get_post_info(args.video, request_base, cookie)
            if post and post.get('post_info'):
                post, set_cookie = post.get('post_info'), post.get('set_cookie')
                username = post.get("author", {}).get("uniqueId") or post.get("author", {}).get("unique_id")
        
                download_dir = username if username and args.as_user else args.download_dir or os.getcwd()
                
                os.makedirs(download_dir, exist_ok=True)
                os.chdir(download_dir)              
                
                index_post = (1, post)               
                download_videos(index_post, args.skip_existing, args.notify, True, args.zip_photos,
                                args.description, args.add_date, cookie, args.add_username, args.convert,
                                args.delete_orig, args.embed_metadata, args.save_metadata, args.keep_corrupt, 1, set_cookie)
   
    if args.loop:
        while True:
            start_scrape()
            time.sleep(args.loop)
    else:
        start_scrape()


if __name__ == '__main__':
    freeze_support()
    main()
